@@ -22,38 +22,36 @@ from gi.repository import GLib
 from gi.repository import Gtk
 
 import meld.ui.util
-from . import dirdiff
-from . import filediff
-from . import filemerge
-from . import melddoc
-from . import newdifftab
-from . import recent
-from . import task
-from . import vcview
-from .ui import gnomeglade
-from .ui import notebooklabel
-
-from .util.compat import string_types
 from meld.conf import _
-from meld.recent import recent_comparisons
+from meld.dirdiff import DirDiff
+from meld.filediff import FileDiff
+from meld.filemerge import FileMerge
+from meld.melddoc import ComparisonState, MeldDoc
+from meld.newdifftab import NewDiffTab
+from meld.recent import recent_comparisons, RecentType
 from meld.settings import interface_settings, settings
+from meld.task import LifoScheduler
+from meld.ui.gnomeglade import Component, ui_file
+from meld.ui.notebooklabel import NotebookLabel
+from meld.vcview import VcView
+from meld.windowstate import SavedWindowState
 
 
-class MeldWindow(gnomeglade.Component):
+class MeldWindow(Component):
 
     def __init__(self):
-        gnomeglade.Component.__init__(self, "meldapp.ui", "meldapp")
+        Component.__init__(self, "meldapp.ui", "meldapp")
         self.widget.set_name("meldapp")
 
         actions = (
             ("FileMenu", None, _("_File")),
-            ("New", Gtk.STOCK_NEW, _("_New Comparison..."), "<Primary>N",
+            ("New", Gtk.STOCK_NEW, _("_New Comparison…"), "<Primary>N",
                 _("Start a new comparison"),
                 self.on_menu_file_new_activate),
             ("Save", Gtk.STOCK_SAVE, None, None,
                 _("Save the current file"),
                 self.on_menu_save_activate),
-            ("SaveAs", Gtk.STOCK_SAVE_AS, _("Save As..."), "<Primary><shift>S",
+            ("SaveAs", Gtk.STOCK_SAVE_AS, _("Save As…"), "<Primary><shift>S",
                 _("Save the current file with a different name"),
                 self.on_menu_save_as_activate),
             ("Close", Gtk.STOCK_CLOSE, None, None,
@@ -73,7 +71,7 @@ class MeldWindow(gnomeglade.Component):
                 self.on_menu_copy_activate),
             ("Paste", Gtk.STOCK_PASTE, None, None, _("Paste the clipboard"),
                 self.on_menu_paste_activate),
-            ("Find", Gtk.STOCK_FIND, _("Find..."), None, _("Search for text"),
+            ("Find", Gtk.STOCK_FIND, _("Find…"), None, _("Search for text"),
                 self.on_menu_find_activate),
             ("FindNext", None, _("Find Ne_xt"), "<Primary>G",
                 _("Search forwards for the same text"),
@@ -82,9 +80,12 @@ class MeldWindow(gnomeglade.Component):
                 _("Search backwards for the same text"),
                 self.on_menu_find_previous_activate),
             ("Replace", Gtk.STOCK_FIND_AND_REPLACE,
-                _("_Replace..."), "<Primary>H",
+                _("_Replace…"), "<Primary>H",
                 _("Find and replace text"),
                 self.on_menu_replace_activate),
+            ("GoToLine", None, _("Go to _Line"), "<Primary>I",
+                _("Go to a specific line"),
+                self.on_menu_go_to_line_activate),
 
             ("ChangesMenu", None, _("_Changes")),
             ("NextChange", Gtk.STOCK_GO_DOWN, _("Next Change"), "<Alt>Down",
@@ -108,22 +109,6 @@ class MeldWindow(gnomeglade.Component):
             ("Refresh", Gtk.STOCK_REFRESH, None, "<Primary>R",
                 _("Refresh the view"),
                 self.on_menu_refresh_activate),
-
-            ("TabMenu", None, _("_Tabs")),
-            ("PrevTab",   None, _("_Previous Tab"), "<Ctrl><Alt>Page_Up",
-                _("Activate previous tab"),
-                self.on_prev_tab),
-            ("NextTab",   None, _("_Next Tab"), "<Ctrl><Alt>Page_Down",
-                _("Activate next tab"),
-                self.on_next_tab),
-            ("MoveTabPrev", None,
-                _("Move Tab _Left"), "<Ctrl><Alt><Shift>Page_Up",
-                _("Move current tab to left"),
-                self.on_move_tab_prev),
-            ("MoveTabNext", None,
-                _("Move Tab _Right"), "<Ctrl><Alt><Shift>Page_Down",
-                _("Move current tab to right"),
-                self.on_move_tab_next),
         )
         toggleactions = (
             ("Fullscreen", None, _("Fullscreen"), "F11",
@@ -133,7 +118,6 @@ class MeldWindow(gnomeglade.Component):
                 _("Show or hide the toolbar"),
                 None, True),
         )
-        ui_file = gnomeglade.ui_file("meldapp-ui.xml")
         self.actiongroup = Gtk.ActionGroup(name='MainActions')
         self.actiongroup.set_translation_domain("meld")
         self.actiongroup.add_actions(actions)
@@ -150,12 +134,12 @@ class MeldWindow(gnomeglade.Component):
 
         self.ui = Gtk.UIManager()
         self.ui.insert_action_group(self.actiongroup, 0)
-        self.ui.add_ui_from_file(ui_file)
+        self.ui.add_ui_from_file(ui_file("meldapp-ui.xml"))
 
         # Manually handle shells that don't show an application menu
         gtk_settings = Gtk.Settings.get_default()
         if not gtk_settings.props.gtk_shell_shows_app_menu:
-            from meldapp import app
+            from meld.meldapp import app
 
             def make_app_action(name):
                 def app_action(*args):
@@ -180,13 +164,9 @@ class MeldWindow(gnomeglade.Component):
             app_actiongroup.add_actions(app_actions)
             self.ui.insert_action_group(app_actiongroup, 0)
 
-            ui_file = gnomeglade.ui_file("appmenu-fallback.xml")
-            self.ui.add_ui_from_file(ui_file)
+            self.ui.add_ui_from_file(ui_file("appmenu-fallback.xml"))
             self.widget.set_show_menubar(False)
 
-        self.tab_switch_actiongroup = None
-        self.tab_switch_merge_id = None
-
         for menuitem in ("Save", "Undo"):
             self.actiongroup.get_action(menuitem).props.is_important = True
         self.widget.add_accel_group(self.ui.get_accel_group())
@@ -225,8 +205,22 @@ class MeldWindow(gnomeglade.Component):
             Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)
         self.toolbar_holder.pack_end(self.secondary_toolbar, False, True, 0)
 
+        # Manually handle GAction additions
+        actions = (
+            ("close", self.on_menu_close_activate, None),
+        )
+        for (name, callback, accel) in actions:
+            action = Gio.SimpleAction.new(name, None)
+            action.connect('activate', callback)
+            self.widget.add_action(action)
+
+        # Create a secondary toolbar, just to hold our progress spinner
         toolbutton = Gtk.ToolItem()
         self.spinner = Gtk.Spinner()
+        # Fake out the spinner on Windows. See Gitlab issue #133.
+        if os.name == 'nt':
+            for attr in ('stop', 'hide', 'show', 'start'):
+                setattr(self.spinner, attr, lambda *args: True)
         toolbutton.add(self.spinner)
         self.secondary_toolbar.insert(toolbutton, -1)
         # Set a minimum size because the spinner requests nothing
@@ -241,15 +235,14 @@ class MeldWindow(gnomeglade.Component):
         self.widget.connect("drag_data_received",
                             self.on_widget_drag_data_received)
 
+        self.window_state = SavedWindowState()
+        self.window_state.bind(self.widget)
+
         self.should_close = False
         self.idle_hooked = 0
-        self.scheduler = task.LifoScheduler()
+        self.scheduler = LifoScheduler()
         self.scheduler.connect("runnable", self.on_scheduler_runnable)
-        window_size = settings.get_value('window-size')
-        self.widget.set_default_size(window_size[0], window_size[1])
-        window_state = settings.get_string('window-state')
-        if window_state == 'maximized':
-            self.widget.maximize()
+
         self.ui.ensure_update()
         self.diff_handler = None
         self.undo_handlers = tuple()
@@ -260,13 +253,9 @@ class MeldWindow(gnomeglade.Component):
         rmenu = self.ui.get_widget('/Menubar/FileMenu/Recent').get_submenu()
         rmenu.connect("map", self._on_recentmenu_map)
 
-        try:
-            builder = meld.ui.util.get_builder("shortcuts.ui")
-            shortcut_window = builder.get_object("shortcuts-meld")
-            self.widget.set_help_overlay(shortcut_window)
-        except GLib.Error:
-            # GtkShortcutsWindow is new in GTK+ 3.20
-            pass
+        builder = meld.ui.util.get_builder("shortcuts.ui")
+        shortcut_window = builder.get_object("shortcuts-meld")
+        self.widget.set_help_overlay(shortcut_window)
 
     def _on_recentmenu_map(self, recentmenu):
         for imagemenuitem in recentmenu.get_children():
@@ -282,16 +271,13 @@ class MeldWindow(gnomeglade.Component):
 
     def on_widget_drag_data_received(self, wid, context, x, y, selection_data,
                                      info, time):
-        if len(selection_data.get_uris()) != 0:
-            paths = []
-            for uri in selection_data.get_uris():
-                paths.append(Gio.File.new_for_uri(uri).get_path())
-            self.open_paths(paths)
+        if len(selection_data.get_files()) != 0:
+            self.open_paths(selection_data.get_files())
             return True
 
     def on_idle(self):
         ret = self.scheduler.iteration()
-        if ret and isinstance(ret, string_types):
+        if ret and isinstance(ret, str):
             self.spinner.set_tooltip_text(ret)
 
         pending = self.scheduler.tasks_pending()
@@ -331,12 +317,6 @@ class MeldWindow(gnomeglade.Component):
 
     def _update_page_action_sensitivity(self):
         current_page = self.notebook.get_current_page()
-        have_prev_tab = current_page > 0
-        have_next_tab = current_page < self.notebook.get_n_pages() - 1
-        self.actiongroup.get_action("PrevTab").set_sensitive(have_prev_tab)
-        self.actiongroup.get_action("NextTab").set_sensitive(have_next_tab)
-        self.actiongroup.get_action("MoveTabPrev").set_sensitive(have_prev_tab)
-        self.actiongroup.get_action("MoveTabNext").set_sensitive(have_next_tab)
 
         if current_page != -1:
             page = self.notebook.get_nth_page(current_page).pyobject
@@ -344,17 +324,17 @@ class MeldWindow(gnomeglade.Component):
             page = None
 
         self.actiongroup.get_action("Close").set_sensitive(bool(page))
-        if not isinstance(page, melddoc.MeldDoc):
+        if not isinstance(page, MeldDoc):
             for action in ("PrevChange", "NextChange", "Cut", "Copy", "Paste",
                            "Find", "FindNext", "FindPrevious", "Replace",
-                           "Refresh"):
+                           "Refresh", "GoToLine"):
                 self.actiongroup.get_action(action).set_sensitive(False)
         else:
             for action in ("Find", "Refresh"):
                 self.actiongroup.get_action(action).set_sensitive(True)
-            is_filediff = isinstance(page, filediff.FileDiff)
+            is_filediff = isinstance(page, FileDiff)
             for action in ("Cut", "Copy", "Paste", "FindNext", "FindPrevious",
-                           "Replace"):
+                           "Replace", "GoToLine"):
                 self.actiongroup.get_action(action).set_sensitive(is_filediff)
 
     def handle_current_doc_switch(self, page):
@@ -387,7 +367,7 @@ class MeldWindow(gnomeglade.Component):
         self.actiongroup.get_action("Redo").set_sensitive(can_redo)
 
         # FileDiff handles save sensitivity; it makes no sense for other modes
-        if not isinstance(newdoc, filediff.FileDiff):
+        if not isinstance(newdoc, FileDiff):
             self.actiongroup.get_action("Save").set_sensitive(False)
             self.actiongroup.get_action("SaveAs").set_sensitive(False)
         else:
@@ -395,12 +375,12 @@ class MeldWindow(gnomeglade.Component):
 
         if newdoc:
             nbl = self.notebook.get_tab_label(newdoc.widget)
-            self.widget.set_title(nbl.get_label_text() + " - Meld")
+            self.widget.set_title(nbl.get_label_text())
             newdoc.on_container_switch_in_event(self.ui)
         else:
             self.widget.set_title("Meld")
 
-        if isinstance(newdoc, melddoc.MeldDoc):
+        if isinstance(newdoc, MeldDoc):
             self.diff_handler = newdoc.connect("next-diff-changed",
                                                self.on_next_diff_changed)
         else:
@@ -410,33 +390,12 @@ class MeldWindow(gnomeglade.Component):
 
     def after_switch_page(self, notebook, page, which):
         self._update_page_action_sensitivity()
-        actiongroup = self.tab_switch_actiongroup
-        if actiongroup:
-            action_name = "SwitchTab%d" % which
-            actiongroup.get_action(action_name).set_active(True)
 
     def after_page_reordered(self, notebook, page, page_num):
         self._update_page_action_sensitivity()
 
-    def on_notebook_label_changed(self, component, text, tooltip):
-        page = component.widget
-        nbl = self.notebook.get_tab_label(page)
-        nbl.set_label_text(text)
-        nbl.set_tooltip_text(tooltip)
-
-        # Only update the window title if the current page is active
-        if self.notebook.get_current_page() == self.notebook.page_num(page):
-            self.widget.set_title(text + " - Meld")
-        if isinstance(text, unicode):
-            text = text.encode('utf8')
-        self.notebook.child_set_property(page, "menu-label", text)
-
-        actiongroup = self.tab_switch_actiongroup
-        if actiongroup:
-            idx = self.notebook.page_num(page)
-            action_name = "SwitchTab%d" % idx
-            label = text.replace("_", "__")
-            actiongroup.get_action(action_name).set_label(label)
+    def on_page_label_changed(self, notebook, label_text):
+        self.widget.set_title(label_text)
 
     def on_can_undo(self, undosequence, can):
         self.actiongroup.get_action("Undo").set_sensitive(can)
@@ -448,17 +407,6 @@ class MeldWindow(gnomeglade.Component):
         self.actiongroup.get_action("PrevChange").set_sensitive(have_prev)
         self.actiongroup.get_action("NextChange").set_sensitive(have_next)
 
-    def on_configure_event(self, window, event):
-        state = event.window.get_state()
-        nosave = Gdk.WindowState.FULLSCREEN | Gdk.WindowState.MAXIMIZED
-        if not (state & nosave):
-            variant = GLib.Variant('(ii)', (event.width, event.height))
-            settings.set_value('window-size', variant)
-
-        maximised = state & Gdk.WindowState.MAXIMIZED
-        window_state = 'maximized' if maximised else 'normal'
-        settings.set_string('window-state', window_state)
-
     def on_menu_file_new_activate(self, menuitem):
         self.append_new_comparison()
 
@@ -505,6 +453,9 @@ class MeldWindow(gnomeglade.Component):
     def on_menu_replace_activate(self, *extra):
         self.current_doc().on_replace_activate()
 
+    def on_menu_go_to_line_activate(self, *extra):
+        self.current_doc().on_go_to_line_activate()
+
     def on_menu_copy_activate(self, *extra):
         widget = self.widget.get_focus()
         if isinstance(widget, Gtk.Editable):
@@ -546,65 +497,6 @@ class MeldWindow(gnomeglade.Component):
     def on_toolbar_stop_clicked(self, *args):
         self.current_doc().stop()
 
-    def on_prev_tab(self, *args):
-        self.notebook.prev_page()
-
-    def on_next_tab(self, *args):
-        self.notebook.next_page()
-
-    def on_move_tab_prev(self, *args):
-        page_num = self.notebook.get_current_page()
-        child = self.notebook.get_nth_page(page_num)
-        page_num = page_num - 1 if page_num > 0 else 0
-        self.notebook.reorder_child(child, page_num)
-
-    def on_move_tab_next(self, *args):
-        page_num = self.notebook.get_current_page()
-        child = self.notebook.get_nth_page(page_num)
-        self.notebook.reorder_child(child, page_num + 1)
-
-    def _update_notebook_menu(self, *args):
-        if self.tab_switch_merge_id:
-            self.ui.remove_ui(self.tab_switch_merge_id)
-            self.ui.remove_action_group(self.tab_switch_actiongroup)
-            self.ui.ensure_update()
-            self.tab_switch_merge_id = None
-            self.tab_switch_actiongroup = None
-
-        if not self.notebook.get_n_pages():
-            return
-
-        self.tab_switch_merge_id = self.ui.new_merge_id()
-        self.tab_switch_actiongroup = Gtk.ActionGroup(name="TabSwitchActions")
-        self.ui.insert_action_group(self.tab_switch_actiongroup)
-        group = None
-        current_page = self.notebook.get_current_page()
-        for i in range(self.notebook.get_n_pages()):
-            page = self.notebook.get_nth_page(i)
-            label = self.notebook.get_menu_label_text(page) or ""
-            label = label.replace("_", "__")
-            name = "SwitchTab%d" % i
-            tooltip = _("Switch to this tab")
-            action = Gtk.RadioAction(
-                name=name, label=label, tooltip=tooltip,
-                stock_id=None, value=i)
-            action.join_group(group)
-            group = action
-            action.set_active(current_page == i)
-
-            def current_tab_changed_cb(action, current):
-                if action == current:
-                    self.notebook.set_current_page(action.get_current_value())
-            action.connect("changed", current_tab_changed_cb)
-            if i < 10:
-                accel = "<Alt>%d" % ((i + 1) % 10)
-            else:
-                accel = None
-            self.tab_switch_actiongroup.add_action_with_accel(action, accel)
-            self.ui.add_ui(self.tab_switch_merge_id,
-                           "/Menubar/TabMenu/TabPlaceholder",
-                           name, name, Gtk.UIManagerItemType.MENUITEM, False)
-
     def page_removed(self, page, status):
         if hasattr(page, 'scheduler'):
             self.scheduler.remove_scheduler(page.scheduler)
@@ -620,6 +512,9 @@ class MeldWindow(gnomeglade.Component):
         if not self.has_pages():
             self.on_switch_page(self.notebook, page, -1)
             self._update_page_action_sensitivity()
+            # Synchronise UIManager state; this shouldn't be necessary,
+            # but upstream aren't touching UIManager bugs.
+            self.ui.ensure_update()
             if self.should_close:
                 cancelled = self.widget.emit(
                     'delete-event', Gdk.Event.new(Gdk.EventType.DELETE))
@@ -627,7 +522,7 @@ class MeldWindow(gnomeglade.Component):
                     self.widget.emit('destroy')
 
     def on_page_state_changed(self, page, old_state, new_state):
-        if self.should_close and old_state == melddoc.STATE_CLOSING:
+        if self.should_close and old_state == ComparisonState.Closing:
             # Cancel closing if one of our tabs does
             self.should_close = False
 
@@ -638,22 +533,20 @@ class MeldWindow(gnomeglade.Component):
                 page.on_file_changed(filename)
 
     def _append_page(self, page, icon):
-        nbl = notebooklabel.NotebookLabel(
-            icon, "", lambda b: page.on_delete_event())
+        nbl = NotebookLabel(icon, "", lambda b: page.on_delete_event())
         self.notebook.append_page(page.widget, nbl)
 
         # Change focus to the newly created page only if the user is on a
         # DirDiff or VcView page, or if it's a new tab page. This prevents
         # cycling through X pages when X diffs are initiated.
-        if isinstance(self.current_doc(), dirdiff.DirDiff) or \
-           isinstance(self.current_doc(), vcview.VcView) or \
-           isinstance(page, newdifftab.NewDiffTab):
+        if isinstance(self.current_doc(), DirDiff) or \
+           isinstance(self.current_doc(), VcView) or \
+           isinstance(page, NewDiffTab):
             self.notebook.set_current_page(self.notebook.page_num(page.widget))
 
         if hasattr(page, 'scheduler'):
             self.scheduler.add_scheduler(page.scheduler)
-        if isinstance(page, melddoc.MeldDoc):
-            page.connect("label-changed", self.on_notebook_label_changed)
+        if isinstance(page, MeldDoc):
             page.connect("file-changed", self.on_file_changed)
             page.connect("create-diff", lambda obj, arg, kwargs:
                          self.append_diff(arg, **kwargs))
@@ -663,9 +556,9 @@ class MeldWindow(gnomeglade.Component):
         self.notebook.set_tab_reorderable(page.widget, True)
 
     def append_new_comparison(self):
-        doc = newdifftab.NewDiffTab(self)
+        doc = NewDiffTab(self)
         self._append_page(doc, "document-new")
-        self.on_notebook_label_changed(doc, _("New comparison"), None)
+        self.notebook.on_label_changed(doc, _("New comparison"), None)
 
         def diff_created_cb(doc, newdoc):
             doc.on_delete_event()
@@ -675,103 +568,111 @@ class MeldWindow(gnomeglade.Component):
         doc.connect("diff-created", diff_created_cb)
         return doc
 
-    def append_dirdiff(self, dirs, auto_compare=False):
+    def append_dirdiff(self, gfiles, auto_compare=False):
+        dirs = [d.get_path() for d in gfiles if d]
         assert len(dirs) in (1, 2, 3)
-        doc = dirdiff.DirDiff(len(dirs))
+        doc = DirDiff(len(dirs))
         self._append_page(doc, "folder")
         doc.set_locations(dirs)
-        # FIXME: This doesn't work, as dirdiff behaves differently to vcview
         if auto_compare:
-            doc.on_button_diff_clicked(None)
+            doc.scheduler.add_task(doc.auto_compare)
         return doc
 
-    def append_filediff(self, files, merge_output=None, meta=None):
-        assert len(files) in (1, 2, 3)
-        doc = filediff.FileDiff(len(files))
+    def append_filediff(
+            self, gfiles, *, encodings=None, merge_output=None, meta=None):
+        assert len(gfiles) in (1, 2, 3)
+        doc = FileDiff(len(gfiles))
         self._append_page(doc, "text-x-generic")
-        doc.set_files(files)
+        doc.set_files(gfiles, encodings)
         if merge_output is not None:
             doc.set_merge_output_file(merge_output)
         if meta is not None:
             doc.set_meta(meta)
         return doc
 
-    def append_filemerge(self, files, merge_output=None):
-        if len(files) != 3:
+    def append_filemerge(self, gfiles, merge_output=None):
+        if len(gfiles) != 3:
             raise ValueError(
-                _("Need three files to auto-merge, got: %r") % files)
-        doc = filemerge.FileMerge(len(files))
+                _("Need three files to auto-merge, got: %r") %
+                [f.get_parse_name() for f in gfiles])
+        doc = FileMerge(len(gfiles))
         self._append_page(doc, "text-x-generic")
-        doc.set_files(files)
+        doc.set_files(gfiles)
         if merge_output is not None:
             doc.set_merge_output_file(merge_output)
         return doc
 
-    def append_diff(self, paths, auto_compare=False, auto_merge=False,
+    def append_diff(self, gfiles, auto_compare=False, auto_merge=False,
                     merge_output=None, meta=None):
-        dirslist = [p for p in paths if os.path.isdir(p)]
-        fileslist = [p for p in paths if os.path.isfile(p)]
-        if dirslist and fileslist:
+        have_directories = False
+        have_files = False
+        for f in gfiles:
+            if f.query_file_type(
+               Gio.FileQueryInfoFlags.NONE, None) == Gio.FileType.DIRECTORY:
+                have_directories = True
+            else:
+                have_files = True
+        if have_directories and have_files:
             raise ValueError(
                 _("Cannot compare a mixture of files and directories"))
-        elif dirslist:
-            return self.append_dirdiff(paths, auto_compare)
+        elif have_directories:
+            return self.append_dirdiff(gfiles, auto_compare)
         elif auto_merge:
-            return self.append_filemerge(paths, merge_output=merge_output)
+            return self.append_filemerge(gfiles, merge_output=merge_output)
         else:
             return self.append_filediff(
-                paths, merge_output=merge_output, meta=meta)
+                gfiles, merge_output=merge_output, meta=meta)
 
     def append_vcview(self, location, auto_compare=False):
-        doc = vcview.VcView()
+        doc = VcView()
         self._append_page(doc, "meld-version-control")
-        location = location[0] if isinstance(location, list) else location
-        doc.set_location(location)
+        if isinstance(location, (list, tuple)):
+            location = location[0]
+        doc.set_location(location.get_path())
         if auto_compare:
-            doc.on_button_diff_clicked(None)
+            doc.scheduler.add_task(doc.auto_compare)
         return doc
 
     def append_recent(self, uri):
-        comparison_type, files, flags = recent_comparisons.read(uri)
-        if comparison_type == recent.TYPE_MERGE:
-            tab = self.append_filemerge(files)
-        elif comparison_type == recent.TYPE_FOLDER:
-            tab = self.append_dirdiff(files)
-        elif comparison_type == recent.TYPE_VC:
-            # Files should be a single-element iterable
-            tab = self.append_vcview(files[0])
-        else:  # comparison_type == recent.TYPE_FILE:
-            tab = self.append_filediff(files)
+        comparison_type, gfiles, flags = recent_comparisons.read(uri)
+        comparison_method = {
+            RecentType.File: self.append_filediff,
+            RecentType.Folder: self.append_dirdiff,
+            RecentType.Merge: self.append_filemerge,
+            RecentType.VersionControl: self.append_vcview,
+        }
+        tab = comparison_method[comparison_type](gfiles)
         self.notebook.set_current_page(self.notebook.page_num(tab.widget))
         recent_comparisons.add(tab)
         return tab
 
-    def _single_file_open(self, path):
-        doc = vcview.VcView()
+    def _single_file_open(self, gfile):
+        doc = VcView()
 
         def cleanup():
             self.scheduler.remove_scheduler(doc.scheduler)
         self.scheduler.add_task(cleanup)
         self.scheduler.add_scheduler(doc.scheduler)
-        path = os.path.abspath(path)
+        path = gfile.get_path()
         doc.set_location(path)
         doc.connect("create-diff", lambda obj, arg, kwargs:
                     self.append_diff(arg, **kwargs))
         doc.run_diff(path)
 
-    def open_paths(self, paths, auto_compare=False, auto_merge=False,
+    def open_paths(self, gfiles, auto_compare=False, auto_merge=False,
                    focus=False):
         tab = None
-        if len(paths) == 1:
-            a = paths[0]
-            if os.path.isfile(a):
-                self._single_file_open(a)
-            else:
+        if len(gfiles) == 1:
+            a = gfiles[0]
+            if a.query_file_type(Gio.FileQueryInfoFlags.NONE, None) == \
+                    Gio.FileType.DIRECTORY:
                 tab = self.append_vcview(a, auto_compare)
+            else:
+                self._single_file_open(a)
 
-        elif len(paths) in (2, 3):
-            tab = self.append_diff(
-                paths, auto_compare=auto_compare, auto_merge=auto_merge)
+        elif len(gfiles) in (2, 3):
+            tab = self.append_diff(gfiles, auto_compare=auto_compare,
+                                   auto_merge=auto_merge)
         if tab:
             recent_comparisons.add(tab)
             if focus:
@@ -785,7 +686,7 @@ class MeldWindow(gnomeglade.Component):
         index = self.notebook.get_current_page()
         if index >= 0:
             page = self.notebook.get_nth_page(index).pyobject
-            if isinstance(page, melddoc.MeldDoc):
+            if isinstance(page, MeldDoc):
                 return page
 
         class DummyDoc(object):
