@@ -16,35 +16,32 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import print_function
-
 import atexit
 import functools
 import logging
-import tempfile
-import shutil
 import os
+import shutil
 import stat
 import sys
+import tempfile
 
 from gi.repository import Gdk
-from gi.repository import GLib
 from gi.repository import Gio
+from gi.repository import GLib
 from gi.repository import GObject
 from gi.repository import Gtk
 from gi.repository import Pango
 
-from meld import melddoc
-from meld import misc
-from meld import recent
 from meld import tree
-from meld import vc
-from meld.ui import gnomeglade
-from meld.ui import vcdialogs
-
 from meld.conf import _
-from meld.settings import settings, bind_settings
-from meld.vc import _null
+from meld.iohelpers import trash_or_confirm
+from meld.melddoc import MeldDoc
+from meld.misc import error_dialog, read_pipe_iter
+from meld.recent import RecentType
+from meld.settings import bind_settings, settings
+from meld.ui.gnomeglade import Component, ui_file
+from meld.ui.vcdialogs import CommitDialog, PushDialog
+from meld.vc import _null, get_vcs
 from meld.vc._vc import Entry
 
 log = logging.getLogger(__name__)
@@ -62,7 +59,7 @@ def cleanup_temp():
             if os.name == "nt":
                 os.chmod(f, stat.S_IWRITE)
             os.remove(f)
-        except:
+        except Exception:
             except_str = "{0[0]}: \"{0[1]}\"".format(sys.exc_info())
             print("File \"{0}\" not removed due to".format(f), except_str,
                   file=sys.stderr)
@@ -71,11 +68,12 @@ def cleanup_temp():
             assert (os.path.exists(f) and os.path.isabs(f) and
                     os.path.dirname(f) == temp_location)
             shutil.rmtree(f, ignore_errors=1)
-        except:
+        except Exception:
             except_str = "{0[0]}: \"{0[1]}\"".format(sys.exc_info())
             print("Directory \"{0}\" not removed due to".format(f), except_str,
                   file=sys.stderr)
 
+
 _temp_dirs, _temp_files = [], []
 atexit.register(cleanup_temp)
 
@@ -120,11 +118,10 @@ class VcTreeStore(tree.DiffTreeStore):
         tree.DiffTreeStore.__init__(self, 1, [str] * 5)
 
     def get_file_path(self, it):
-        # Use instead of value_path; does not incorrectly decode
         return self.get_value(it, self.column_index(tree.COL_PATH, 0))
 
 
-class VcView(melddoc.MeldDoc, gnomeglade.Component):
+class VcView(MeldDoc, Component):
 
     __gtype_name__ = "VcView"
 
@@ -155,20 +152,13 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         "ignored": ("VcShowIgnored", Entry.is_ignored),
     }
 
-    file_encoding = sys.getfilesystemencoding()
-
-    @classmethod
-    def display_path(cls, bytes):
-        encodings = (cls.file_encoding,)
-        return misc.fallback_decode(bytes, encodings, lossy=True)
-
     def __init__(self):
-        melddoc.MeldDoc.__init__(self)
-        gnomeglade.Component.__init__(self, "vcview.ui", "vcview",
-                                      ["VcviewActions", 'liststore_vcs'])
+        MeldDoc.__init__(self)
+        Component.__init__(
+            self, "vcview.ui", "vcview", ["VcviewActions", 'liststore_vcs'])
         bind_settings(self)
 
-        self.ui_file = gnomeglade.ui_file("vcview-ui.xml")
+        self.ui_file = ui_file("vcview-ui.xml")
         self.actiongroup = self.VcviewActions
         self.actiongroup.set_translation_domain("meld")
         self.model = VcTreeStore()
@@ -177,8 +167,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self.treeview.set_model(self.model)
         self.treeview.get_selection().connect(
             "changed", self.on_treeview_selection_changed)
-        self.treeview.set_search_equal_func(
-            self.model.treeview_search_cb, None)
+        self.treeview.set_search_equal_func(tree.treeview_search_cb, None)
         self.current_path, self.prev_path, self.next_path = None, None, None
 
         self.name_column.set_attributes(
@@ -251,7 +240,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             location = parent_location
         else:
             # existing parent directory was found
-            for avc in vc.get_vcs(location):
+            for avc in get_vcs(location):
                 err_str = ''
                 vc_details = {'name': avc.NAME, 'cmd': avc.CMD}
 
@@ -321,6 +310,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             return
 
         root = self.model.get_iter_first()
+        root_path = self.model.get_path(root)
 
         try:
             self.model.set_value(
@@ -329,23 +319,39 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             pass
 
         self.scheduler.add_task(self.vc.refresh_vc_state)
-        self.scheduler.add_task(self._search_recursively_iter(root))
+        self.scheduler.add_task(self._search_recursively_iter(root_path))
         self.scheduler.add_task(self.on_treeview_selection_changed)
         self.scheduler.add_task(self.on_treeview_cursor_changed)
 
     def get_comparison(self):
-        return recent.TYPE_VC, [self.location]
+        uris = [Gio.File.new_for_path(self.location)]
+        return RecentType.VersionControl, uris
 
     def recompute_label(self):
-        location = self.display_path(self.location)
-        self.label_text = os.path.basename(location)
+        self.label_text = os.path.basename(self.location)
         # TRANSLATORS: This is the location of the directory being viewed
-        self.tooltip_text = _("%s: %s") % (_("Location"), location)
+        self.tooltip_text = _("%s: %s") % (_("Location"), self.location)
         self.label_changed()
 
-    def _search_recursively_iter(self, iterstart):
+    def _search_recursively_iter(self, start_path, replace=False):
+
+        # Initial yield so when we add this to our tasks, we don't
+        # create iterators that may be invalidated.
+        yield _("Scanning repository")
+
+        if replace:
+            # Replace the row at start_path with a new, empty row ready
+            # to be filled.
+            old_iter = self.model.get_iter(start_path)
+            file_path = self.model.get_file_path(old_iter)
+            new_iter = self.model.insert_after(None, old_iter)
+            self.model.set_value(new_iter, tree.COL_PATH, file_path)
+            self.model.set_path_state(new_iter, 0, tree.STATE_NORMAL, True)
+            self.model.remove(old_iter)
+
+        iterstart = self.model.get_iter(start_path)
         rootname = self.model.get_file_path(iterstart)
-        display_prefix = len(self.display_path(rootname)) + 1
+        display_prefix = len(rootname) + 1
         symlinks_followed = set()
         todo = [(self.model.get_path(iterstart), rootname)]
 
@@ -360,7 +366,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             todo.sort()
             treepath, path = todo.pop(0)
             it = self.model.get_iter(treepath)
-            yield _("Scanning %s") % self.display_path(path)[display_prefix:]
+            yield _("Scanning %s") % path[display_prefix:]
 
             entries = self.vc.get_entries(path)
             entries = [e for e in entries if any(f(e) for f in filters)]
@@ -430,10 +436,10 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def run_diff(self, path):
         if os.path.isdir(path):
-            self.emit("create-diff", [path], {})
+            self.emit("create-diff", [Gio.File.new_for_path(path)], {})
             return
 
-        basename = self.display_path(os.path.basename(path))
+        basename = os.path.basename(path)
         meta = {
             'parent': self,
             'prompt_resolve': False,
@@ -486,7 +492,8 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             os.chmod(temp_file, 0o444)
             _temp_files.append(temp_file)
 
-        self.emit("create-diff", diffs, kwargs)
+        self.emit("create-diff",
+                  [Gio.File.new_for_path(d) for d in diffs], kwargs)
 
     def do_popup_treeview_menu(self, widget, event):
         if event:
@@ -520,9 +527,10 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         return False
 
     def on_filter_state_toggled(self, button):
-        active_action = lambda a: self.actiongroup.get_action(a).get_active()
         active_filters = [
-            k for k, v in self.state_actions.items() if active_action(v[0])]
+            k for k, (action_name, fn) in self.state_actions.items()
+            if self.actiongroup.get_action(action_name).get_active()
+        ]
 
         if set(active_filters) == set(self.props.status_filters):
             return
@@ -575,7 +583,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         files = [os.path.relpath(f, working_dir) for f in files]
         msg = shelljoin(command + files) + " (in %s)\n" % working_dir
         self.consolestream.command(msg)
-        readiter = misc.read_pipe_iter(
+        readiter = read_pipe_iter(
             command + files, workdir=working_dir,
             errorstream=self.consolestream)
         try:
@@ -584,7 +592,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 yield 1
                 result = next(readiter)
         except IOError as err:
-            misc.error_dialog(
+            error_dialog(
                 "Error running command",
                 "While running '%s'\nError: %s" % (msg, err))
             result = (1, "")
@@ -603,7 +611,20 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         vc_command = self.command_map.get(command)
         return vc_command and hasattr(self.vc, vc_command)
 
-    def command(self, command, files):
+    def command(self, command, files, sync=False):
+        """
+        Run a command against this view's version control subsystem
+
+        This is the intended way for things outside of the VCView to
+        call in to version control methods, e.g., to mark a conflict as
+        resolved from a file comparison.
+
+        :param command: The version control command to run, taken from
+            keys in `VCView.command_map`.
+        :param files: File parameters to the command as paths
+        :param sync: If True, the command will be executed immediately
+            (as opposed to being run by the idle scheduler).
+        """
         if not self.has_command(command):
             log.error("Couldn't understand command %s", command)
             return
@@ -612,24 +633,30 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             log.error("Invalid files argument to '%s': %r", command, files)
             return
 
+        runner = self.runner if not sync else self.sync_runner
         command = getattr(self.vc, self.command_map[command])
-        command(self.runner, files)
+        command(runner, files)
 
     def runner(self, command, files, refresh, working_dir):
         """Schedule a version control command to run as an idle task"""
         self.scheduler.add_task(
             self._command_iter(command, files, refresh, working_dir))
 
+    def sync_runner(self, command, files, refresh, working_dir):
+        """Run a version control command immediately"""
+        for it in self._command_iter(command, files, refresh, working_dir):
+            pass
+
     def on_button_update_clicked(self, obj):
         self.vc.update(self.runner)
 
     def on_button_push_clicked(self, obj):
-        response = vcdialogs.PushDialog(self).run()
+        response = PushDialog(self).run()
         if response == Gtk.ResponseType.OK:
             self.vc.push(self.runner)
 
     def on_button_commit_clicked(self, obj):
-        response, commit_msg = vcdialogs.CommitDialog(self).run()
+        response, commit_msg = CommitDialog(self).run()
         if response == Gtk.ResponseType.OK:
             self.vc.commit(
                 self.runner, self._get_selected_files(), commit_msg)
@@ -669,11 +696,17 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def on_button_delete_clicked(self, obj):
         files = self._get_selected_files()
         for name in files:
+            gfile = Gio.File.new_for_path(name)
+
             try:
-                gfile = Gio.File.new_for_path(name)
-                gfile.trash(None)
-            except GLib.GError as e:
-                misc.error_dialog(_("Error removing %s") % name, str(e))
+                trash_or_confirm(gfile)
+            except Exception as e:
+                error_dialog(
+                    _("Error deleting {}").format(
+                        GLib.markup_escape_text(gfile.get_parse_name()),
+                    ),
+                    str(e),
+                )
 
         workdir = os.path.dirname(os.path.commonprefix(files))
         self.refresh_partial(workdir)
@@ -695,17 +728,16 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def refresh_partial(self, where):
         if not self.actiongroup.get_action("VcFlatten").get_active():
             it = self.find_iter_by_name(where)
-            if it:
-                newiter = self.model.insert_after(None, it)
-                self.model.set_value(
-                    newiter, tree.COL_PATH, where)
-                self.model.set_path_state(newiter, 0, tree.STATE_NORMAL, True)
-                self.model.remove(it)
-                self.treeview.grab_focus()
-                self.treeview.get_selection().select_iter(newiter)
-                self.scheduler.add_task(self._search_recursively_iter(newiter))
-                self.scheduler.add_task(self.on_treeview_selection_changed)
-                self.scheduler.add_task(self.on_treeview_cursor_changed)
+            if not it:
+                return
+            path = self.model.get_path(it)
+
+            self.treeview.grab_focus()
+            self.vc.refresh_vc_state(where)
+            self.scheduler.add_task(
+                self._search_recursively_iter(path, replace=True))
+            self.scheduler.add_task(self.on_treeview_selection_changed)
+            self.scheduler.add_task(self.on_treeview_cursor_changed)
         else:
             # XXX fixme
             self.refresh()
@@ -752,9 +784,9 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_consoleview_populate_popup(self, textview, menu):
         buf = textview.get_buffer()
-        clear_cb = lambda *args: buf.delete(*buf.get_bounds())
         clear_action = Gtk.MenuItem.new_with_label(_("Clear"))
-        clear_action.connect("activate", clear_cb)
+        clear_action.connect(
+            "activate", lambda *args: buf.delete(*buf.get_bounds()))
         menu.insert(clear_action, 0)
         menu.insert(Gtk.SeparatorMenuItem(), 1)
         menu.show_all()
@@ -812,3 +844,10 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_find_activate(self, *extra):
         self.treeview.emit("start-interactive-search")
+
+    def auto_compare(self):
+        modified_states = (tree.STATE_MODIFIED, tree.STATE_CONFLICT)
+        for it in self.model.state_rows(modified_states):
+            row_paths = self.model.value_paths(it)
+            paths = [p for p in row_paths if os.path.exists(p)]
+            self.run_diff(paths[0])
